#!/usr/bin/env python
#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Default config for legion
"""

LEGION_API_ADDR = '0.0.0.0'
LEGION_API_PORT = 5001

NAMESPACE = ''

DEBUG = False

AUTH_ENABLED = False

AUTH_TOKEN_ENABLED = True
AUTH_TOKEN = 'demo-token'

REGISTER_ON_GRAFANA = True

CLUSTER_CONFIG_PATH = '/opt/legion/state/cluster.yaml'
CLUSTER_SECRETS_PATH = '/opt/legion/secrets'
JWT_CONFIG_PATH = '/opt/legion/jwtconf'



#
#    Copyright 2018 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
legion k8s definitions functions
"""
import typing

LEGION_SYSTEM_LABEL = 'legion.system'
LEGION_SYSTEM_VALUE = 'yes'
LEGION_COMPONENT_LABEL = 'legion.component'
LEGION_COMPONENT_NAME_MODEL = 'model'
LEGION_COMPONENT_NAME_EDI = 'edi'
LEGION_COMPONENT_NAME_API = 'edge'
LEGION_COMPONENT_NAME_GRAFANA = 'grafana'
LEGION_COMPONENT_NAME_GRAPHITE = 'graphite'

LEGION_API_SERVICE_PORT = 'api'

ENCLAVE_NAMESPACE_LABEL = 'enclave'

INGRESS_DASHBOARD_NAME = 'kubernetes-dashboard'
INGRESS_GRAFANA_NAME = 'legion-core-grafana'

STATUS_OK = 'ok'
STATUS_FAIL = 'fail'
STATUS_WARN = 'warning'

EVENT_ADDED = 'ADDED'
EVENT_MODIFIED = 'MODIFIED'
EVENT_DELETED = 'DELETED'

LOAD_DATA_ITERATIONS = 5
LOAD_DATA_TIMEOUT = 2

ModelContainerMetaInformation = typing.NamedTuple('ModelContainerMetaInformation', [
    ('k8s_name', str),
    ('model_id', str),
    ('model_version', str),
    ('kubernetes_labels', typing.Dict[str, str]),
    ('kubernetes_annotations', typing.Dict[str, str]),
])


class ModelIdVersion:
    """
    Holder for model ID and version
    """

    def __init__(self, id, version):
        """
        Build model ID and version holder

        :param id: model ID
        :type id: str
        :param version: model version
        :type version: str
        """
        self._id = id
        self._version = version

    @property
    def id(self):
        """
        Get model ID

        :return: str -- model ID
        """
        return self._id

    @property
    def version(self):
        """
        Get model version

        :return: str -- model version
        """
        return self._version

    def __eq__(self, other):
        """
        Check equation of self object with another by fields

        :param other: another object
        :type other: :py:class:`legion.k8s.definitions.ModelIdVersion`
        :return: bool -- result of equation
        """
        return self.id == other.id and self.version == other.version

    def __repr__(self):
        """
        Get str representation for repr function

        :return: str -- string representation
        """
        return '{}({}, {})'.format(self.__class__.__name__,
                                   self.id, self.version)

    def __hash__(self):
        """
        Build hash of object

        :return: int -- hash of object
        """
        return hash((self.id, self.version))

    __str__ = __repr__


class ModelDeploymentDescription:
    """
    Holder for model deployment description
    """

    def __init__(self,
                 status,
                 model, version,
                 image,
                 scale, ready_replicas,
                 namespace,
                 model_api_ok=None, model_api_info=None):
        """
        Construct model deployment description

        :param status: status of model deployment
        :type status: str
        :param model: model ID
        :type model: str
        :param version: model version
        :type version: str
        :param image: model image
        :type image: str
        :param scale: model scale
        :type scale: int
        :param ready_replicas: already ready model replicas
        :type ready_replicas: int
        :param namespace: model namespace
        :type namespace: str
        :param model_api_ok: (Optional) model API state from EDGE invocation
        :type model_api_ok: bool or None
        :param model_api_info: (Optional) model API info from EDGE invocation
        :type model_api_info: dict or None
        """
        self._status = status
        self._model = model
        self._version = version
        self._image = image
        self._scale = scale
        self._ready_replicas = ready_replicas
        self._namespace = namespace
        self._model_api_ok = model_api_ok
        self._model_api_info = model_api_info

    @staticmethod
    def build_from_model_service(model_service, model_api_ok=None, model_api_info=None):
        """
        Build from existed model service

        :param model_service: model service
        :type model_service: :py:class:`legion.k8s.services.ModelService`
        :param model_api_ok: (Optional) model API state from EDGE invocation
        :type model_api_ok: bool or None
        :param model_api_info: (Optional) model API info from EDGE invocation
        :type model_api_info: dict or None
        :return: :py:class:`legion.k8s.definitions.ModelDeploymentDescription` -- built model deployment
        """
        return ModelDeploymentDescription(
            status=model_service.status,
            model=model_service.id,
            version=model_service.version,
            image=model_service.image,
            scale=model_service.desired_scale,
            ready_replicas=model_service.scale,
            namespace=model_service.namespace,
            model_api_ok=model_api_ok,
            model_api_info=model_api_info,
        )

    @property
    def id_and_version(self):
        """
        Build key from model id and model version

        :return: :py:class:`legion.k8s.definitions.ModelIdVersion`
        """
        return ModelIdVersion(self.model, self.version)

    @property
    def status(self):
        """
        Get model status

        :return: str -- model status [STATUS_OK, STATUS_FAIL, STATUS_WARN]
        """
        return self._status

    @property
    def model(self):
        """
        Get model ID

        :return: str -- model ID
        """
        return self._model

    @property
    def version(self):
        """
        Get model version

        :return: str -- model version
        """
        return self._version

    @property
    def image(self):
        """
        Get model image

        :return: str -- model image
        """
        return self._image

    @property
    def scale(self):
        """
        Get model scale

        :return: int -- model scale
        """
        return self._scale

    @property
    def ready_replicas(self):
        """
        Get actual ready replicas

        :return: int -- actual ready replicas
        """
        return self._ready_replicas

    @property
    def namespace(self):
        """
        Get model namespace

        :return: str -- model namespace
        """
        return self._namespace

    @property
    def model_api_ok(self):
        """
        Get model API status from model invocation

        :return: str or None -- model API status
        """
        return self._model_api_ok

    @property
    def model_api_info(self):
        """
        Get info about model from model API

        :return: dict -- model info
        """
        return self._model_api_info

    def as_dict(self):
        """
        Get current object as a JSON-serializable dict

        :return: dict[str, Any] -- current object status
        """
        return {
            'status': self._status,
            'model': self._model,
            'version': self._version,
            'image': self._image,
            'scale': self._scale,
            'ready_replicas': self._ready_replicas,
            'namespace': self._namespace,
            'model_api_ok': self._model_api_ok,
            'model_api_info': self._model_api_info
        }

    @staticmethod
    def build_from_json(json_dict):
        """
        Build ModelDeploymentDescription from JSON dict

        :param json_dict: result of deserialization serialized .as_dict response
        :type json_dict: dict[str, Any]
        :return: :py:class:`legion.k8s.definitions.ModelDeploymentDescription` -- built model deployment
        """
        if isinstance(json_dict['model_api_info'], bytes):
            json_dict['model_api_info'] = json.loads(json_dict['model_api_info'])
        return ModelDeploymentDescription(**json_dict)

    def __repr__(self):
        """
        Get string representation of object

        :return: str -- string representation
        """
        args = ",".join(repr(arg) for arg in [
            self.status, self.model, self.version, self.image,
            self.scale, self.ready_replicas, self.namespace,
            self.model_api_ok, self.model_api_info
        ])
        return "ModelDeploymentDescription({})".format(args)

    __str__ = __repr__



#
#    Copyright 2018 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
legion k8s watch context manager
"""
import logging

import kubernetes
import kubernetes.client.rest
import kubernetes.client
import kubernetes.config
import kubernetes.config.config_exception
import urllib3
import urllib3.exceptions

LOGGER = logging.getLogger(__name__)
CONNECTION_CONTEXT = None


class ResourceWatch:
    """
    Watch for K8S resources (context manager)
    """

    def __init__(self, api_function, *args,
                 filter_callable=None,
                 object_constructor=None,
                 resource_version=None,
                 **kwargs):
        """
        Initialize context manager for resource watch

        :param api_function: API function to watch
        :param args: additional positional arguments for API function
        :param filter_callable: (Optional) callable to filter objects
        :param object_constructor:  (Optional) callable to construct object wrappers
        :param resource_version: (Optional) start resource version
        :param kwargs: additional key value arguments for API function
        """
        self._api_function = api_function
        self._filter_callable = filter_callable
        self._object_constructor = object_constructor
        self._resource_version = resource_version
        self._args = args
        self._kwargs = kwargs

    @property
    def stream(self):
        """
        Access watch stream

        :return: tuple(str, Any) -- event type and event object (plain or constructed if
                                    object_constructor has been passed)
        """
        LOGGER.debug('Starting watch stream')

        while True:
            LOGGER.debug('Entering event loop iteration in watch')
            try:
                watch = kubernetes.watch.Watch()
                kwargs = self._kwargs
                if self._resource_version:
                    LOGGER.debug('Using latest resource version: {}'.format(self._resource_version))
                    kwargs['resource_version'] = self._resource_version

                LOGGER.debug('Creating watcher for function {}. Args: {!r}, Kwargs: {!r}'
                             .format(self._api_function.__name__, self._args, kwargs))

                stream = watch.stream(self._api_function,
                                      *self._args,
                                      **kwargs)
                for event in stream:
                    event_type = event['type']
                    event_object = event['object']

                    if hasattr(event['object'], 'metadata') and hasattr(event['object'].metadata, 'resource_version'):
                        self._resource_version = event['object'].metadata.resource_version

                    # Check if valid object
                    pass_event = not self._filter_callable or self._filter_callable(event_object)

                    if pass_event:
                        # Construct specific object
                        if self._object_constructor:
                            event_object = self._object_constructor(event_object)

                        yield (event_type, event_object)
            except urllib3.exceptions.ProtocolError as protocol_error:
                LOGGER.warning('Connection to K8S API has been lost: {}. Reconnecting...'.format(protocol_error))
                continue
            except kubernetes.client.rest.ApiException as kube_api_exception:
                LOGGER.warning('Got Kubernetes API exception: {}'.format(kube_api_exception))
                if kube_api_exception.status == 500:
                    LOGGER.warning('Got Kubernetes error 500. Reconnecting...')
                    continue
                else:
                    LOGGER.warning('Got wrong error code. Breaking...')
                    break
            except Exception as general_exception:
                LOGGER.exception('Got general exception: {}. Breaking...'.format(general_exception))
                break

        LOGGER.debug('Watch stream has been ended')



#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
EDI client
"""

import json
import logging
import os

import legion.k8s
import legion.edi.server
import legion.config
import requests
import requests.exceptions

LOGGER = logging.getLogger(__name__)


class EdiClient:
    """
    EDI client
    """

    def __init__(self, base, user=None, password=None, token=None, retries=3):
        """
        Build client

        :param base: base url, for example: http://edi.parallels
        :type base: str
        :param user: user name for user/password based auth
        :type user: str or None
        :param password: user password for user/password based auth
        :type password: str or None
        :param token: token for token based auth
        :type token: str or None
        :param retries: command retries or less then 2 if disabled
        :type retries: int
        """
        self._base = base
        self._user = user
        self._password = password
        self._token = token
        self._version = legion.edi.server.EDI_VERSION
        self._retries = retries

    def _query(self, url_template, payload=None, action='GET'):
        """
        Perform query to EDI server

        :param url_template: url template from legion.const.api
        :type url_template: str
        :param payload: payload (will be converted to JSON) or None
        :type payload: dict[str, any]
        :param action: HTTP method (GET, POST, PUT, DELETE)
        :type action: str
        :return: dict[str, any] -- response content
        """
        sub_url = url_template.format(version=self._version)
        full_url = self._base.strip('/') + sub_url

        auth = None
        headers = {}

        if self._user and self._password:
            auth = (self._user, self._password)
        elif self._token:
            auth = ('token', self._token)

        left_retries = self._retries if self._retries > 0 else 1
        raised_exception = None

        while left_retries > 0:
            try:
                LOGGER.debug('Requesting {}'.format(full_url))
                response = requests.request(action.lower(), full_url, data=payload, headers=headers, auth=auth)
            except requests.exceptions.ConnectionError as exception:
                LOGGER.error('Failed to connect to {}: {}. Retrying'.format(self._base, exception))
                raised_exception = exception
            else:
                LOGGER.debug('Got response. Breaking')
                break

            left_retries -= 1
        else:
            raise Exception('Failed to connect to {}. No one retry left. Exception: {}'.format(
                self._base, raised_exception
            ))

        try:
            answer = json.loads(response.text)
            LOGGER.debug('Got answer: {!r} with code {} for URL {!r}'
                         .format(answer, response.status_code, full_url))
        except ValueError as json_decode_exception:
            raise ValueError('Invalid JSON structure {!r}: {}'.format(response.text, json_decode_exception))

        if isinstance(answer, dict) and answer.get('error', False):
            exception = answer.get('exception')
            raise Exception('Got error from server: {!r}'.format(exception))

        if response.status_code != 200:
            raise Exception('Server returned wrong HTTP code (not 200) without error flag')

        LOGGER.debug('Query has been completed, parsed and validated')
        return answer

    @staticmethod
    def parse_deployments(response):
        """
        Parse model deployments from response

        :param response: EDI server response
        :type response: list[dict[str, any]]
        :return: list[:py:class:`legion.k8s.ModelDeploymentDescription`] -- parsed model deployments
        """
        return [legion.k8s.ModelDeploymentDescription.build_from_json(x) for x in response]

    def inspect(self, model=None, version=None):
        """
        Perform inspect query on EDI server

        :param model: model id
        :type model: str
        :param version: (Optional) model version
        :type version: str
        :return: list[:py:class:`legion.containers.k8s.ModelDeploymentDescription`]
        """
        payload = {}
        if model:
            payload['model'] = model
        if version:
            payload['version'] = version

        return self.parse_deployments(self._query(legion.edi.server.EDI_INSPECT, payload=payload))

    def info(self):
        """
        Perform info query on EDI server

        :return: dict[:py:class:`legion.containers.k8s.ModelDeploymentDescription`]
        """
        return self._query(legion.edi.server.EDI_INFO)

    def deploy(self, image, count=1, livenesstimeout=2, readinesstimeout=2):
        """
        Deploy API endpoint

        :param image: Docker image for deploy (for kubernetes deployment and local pull)
        :type image: str
        :param count: count of pods to create
        :type count: int
        :param livenesstimeout: model pod startup timeout (used in liveness probe)
        :type livenesstimeout: int
        :param readinesstimeout: model pod startup timeout (used readiness probe)
        :type readinesstimeout: int
        :return: list[:py:class:`legion.containers.k8s.ModelDeploymentDescription`] -- affected model deployments
        """
        payload = {
            'image': image
        }
        if count is not None:
            payload['count'] = count
        if livenesstimeout:
            payload['livenesstimeout'] = livenesstimeout
        if readinesstimeout:
            payload['readinesstimeout'] = readinesstimeout

        return self.parse_deployments(self._query(legion.edi.server.EDI_DEPLOY, action='POST', payload=payload))

    def undeploy(self, model, grace_period=0, version=None, ignore_not_found=False):
        """
        Undeploy API endpoint

        :param model: model id
        :type model: str
        :param grace_period: grace period for removing
        :type grace_period: int
        :param version: (Optional) model version
        :type version: str
        :param ignore_not_found: (Optional) ignore if cannot find models
        :type ignore_not_found: bool
        :return: list[:py:class:`legion.containers.k8s.ModelDeploymentDescription`] -- affected model deployments
        """
        payload = {
            'model': model,
            'ignore_not_found': ignore_not_found
        }

        if grace_period:
            payload['grace_period'] = grace_period

        if version:
            payload['version'] = version

        return self.parse_deployments(self._query(legion.edi.server.EDI_UNDEPLOY, action='POST', payload=payload))

    def scale(self, model, count, version=None):
        """
        Scale model

        :param model: model id
        :type model: str
        :param count: count of pods to create
        :type count: int
        :param version: (Optional) model version
        :type version: str
        :return: list[:py:class:`legion.containers.k8s.ModelDeploymentDescription`] -- affected model deployments
        """
        payload = {
            'model': model,
            'count': count
        }
        if version:
            payload['version'] = version

        return self.parse_deployments(self._query(legion.edi.server.EDI_SCALE, action='POST', payload=payload))

    def get_token(self, version=None):
        """
        Get API token

        :param version: (Optional) model version
        :type version: str
        :return: str -- return API Token
        """
        payload = {}
        if version:
            payload['version'] = version

        response = self._query(legion.edi.server.EDI_GENERATE_TOKEN, action='GET', payload=payload)
        if response and 'token' in response:
            return response['token']

    def __repr__(self):
        """
        Get string representation of object

        :return: str -- string representation
        """
        return "EdiClient({!r})".format(self._base)

    __str__ = __repr__


def add_edi_arguments(parser):
    """
    Add EDI arguments parser

    :param parser:
    :type parser:
    :return:
    """
    parser.add_argument('--edi',
                        type=str, help='EDI server host')
    parser.add_argument('--user',
                        type=str, help='EDI server user')
    parser.add_argument('--password',
                        type=str, help='EDI server password')
    parser.add_argument('--token',
                        type=str, help='EDI server token')


def build_client(args=None):
    """
    Build EDI client from from ENV and from command line arguments

    :param args: (optional) command arguments with .namespace
    :type args: :py:class:`argparse.Namespace`
    :return: :py:class:`legion.external.edi.EdiClient` -- EDI client
    """
    host, user, password, token = None, None, None, None

    if args:
        if args.edi:
            host = args.edi

        if args.user:
            user = args.user

        if args.password:
            password = args.password

        if args.token:
            token = args.token

    if not host:
        try:
            host = legion.k8s.Enclave(os.environ.get("NAMESPACE")).edi_service.url
        except Exception:
            host = os.environ.get(*legion.config.EDI_URL)

    if not user or not password:
        user = os.environ.get(*legion.config.EDI_USER)
        password = os.environ.get(*legion.config.EDI_PASSWORD)

    if not token:
        token = os.environ.get(*legion.config.EDI_TOKEN)

    client = EdiClient(host, user, password, token)
    return client



#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Flask app
"""

import logging
import os
import itertools

import legion.config
import legion.http
import legion.model
import legion.pymodel
import legion.k8s.properties
import legion.k8s.utils
from flask import Flask, Blueprint, request, jsonify, redirect, render_template
from flask import current_app as app

LOGGER = logging.getLogger(__name__)
blueprint = Blueprint('pyserve', __name__)

SERVE_ROOT = '/'
SERVE_INFO = '/api/model/{model_id}/{model_version}/info'
SERVE_INVOKE = '/api/model/{model_id}/{model_version}/invoke/{endpoint}'
SERVE_INVOKE_DEFAULT = '/api/model/{model_id}/{model_version}/invoke'
SERVE_BATCH = '/api/model/{model_id}/{model_version}/batch/{endpoint}'
SERVE_BATCH_DEFAULT = '/api/model/{model_id}/{model_version}/batch'
SERVE_PROPERTIES = '/api/model/{model_id}/{model_version}/properties'
SERVE_EMIT_PROPERTIES = '/api/model/{model_id}/{model_version}/emit-properties-update'
SERVE_HEALTH_CHECK = '/healthcheck'

ALL_URLS = SERVE_ROOT, \
           SERVE_INFO, \
           SERVE_INVOKE, SERVE_INVOKE_DEFAULT, \
           SERVE_BATCH, SERVE_BATCH_DEFAULT, \
           SERVE_PROPERTIES, SERVE_EMIT_PROPERTIES, \
           SERVE_HEALTH_CHECK


def validate_model_id(model_id, model_version):
    """
    Check that passed model info is equal to current model info

    :param model_id: model id
    :type model_id: str
    :param model_version: model version
    :type model_version: str
    :return: None
    """
    if model_id != app.config['model'].model_id:
        raise Exception('Invalid model handler: {}, not {}'.format(app.config['model'].model_id, model_id))
    if model_version != app.config['model'].model_version:
        raise Exception('Invalid model handler: {}, not {}'.format(app.config['model'].model_version, model_version))


@blueprint.route(SERVE_ROOT)
def root():
    """
    Return static file for root query

    :return: :py:class:`Flask.Response` -- response index file
    """
    return redirect('index.html')


@blueprint.route(SERVE_INFO.format(model_id='<model_id>', model_version='<model_version>'))
def model_info(model_id, model_version):
    """
    Get model description

    :param model_id: model id
    :type model_id: str
    :param model_version: model version
    :type model_version: str
    :return: :py:class:`Flask.Response` -- model description
    """
    validate_model_id(model_id, model_version)

    model = app.config['model']

    return jsonify(model.description)


@blueprint.route(SERVE_INVOKE_DEFAULT.format(model_id='<model_id>', model_version='<model_version>'),
                 methods=['POST', 'GET'])
@blueprint.route(SERVE_INVOKE.format(model_id='<model_id>', model_version='<model_version>',
                                     endpoint='<endpoint>'), methods=['POST', 'GET'])
def model_invoke(model_id, model_version, endpoint='default'):
    """
    Call model for calculation

    :param model_id: model name
    :type model_id: str
    :param model_version: model version
    :type model_version: str
    :param endpoint: target endpoint name
    :type endpoint: str
    :return: :py:class:`Flask.Response` -- result of calculation
    """
    validate_model_id(model_id, model_version)

    input_dict = legion.http.parse_request(request)

    model = app.config['model']
    if endpoint not in model.endpoints:
        raise Exception('Unknown endpoint {!r}'.format(endpoint))

    output = model.endpoints[endpoint].invoke(input_dict)

    return legion.http.prepare_response(output)


@blueprint.route(SERVE_BATCH_DEFAULT.format(model_id='<model_id>', model_version='<model_version>'),
                 methods=['POST'])
@blueprint.route(SERVE_BATCH.format(model_id='<model_id>', model_version='<model_version>', endpoint='<endpoint>'),
                 methods=['POST'])
def model_batch(model_id, model_version, endpoint='default'):
    """
    Call model for calculation in batch mode

    :param model_id: model name
    :type model_id: str
    :param model_version: model version
    :type model_version: str
    :param endpoint: target endpoint name
    :type endpoint: str
    :return: :py:class:`Flask.Response` -- result of calculation
    """
    validate_model_id(model_id, model_version)

    input_dicts = legion.http.parse_batch_request(request)

    model = app.config['model']
    if endpoint not in model.endpoints:
        raise Exception('Unknown endpoint {!r}'.format(endpoint))

    responses = [model.endpoints[endpoint].invoke(input_dict) for input_dict in input_dicts]

    return legion.http.prepare_response(responses)


@blueprint.route(SERVE_HEALTH_CHECK)
def healthcheck():
    """
    Check that model is OK

    :return: str -- status string
    """
    return 'OK'


@blueprint.route(SERVE_PROPERTIES.format(model_id='<model_id>', model_version='<model_version>'))
def model_properties(model_id, model_version):
    """
    Get model properties

    :param model_id: model id
    :type model_id: str
    :param model_version: model version
    :type model_version: str
    :return: :py:class:`Flask.Response` -- model properties
    """
    validate_model_id(model_id, model_version)

    model = app.config['model']

    return jsonify(model.properties.data)


@blueprint.route(SERVE_EMIT_PROPERTIES.format(model_id='<model_id>', model_version='<model_version>'))
def emit_properties_update_signal(model_id, model_version):
    """
    Emit signal for properties update

    :param model_id: model id
    :type model_id: str
    :param model_version: model version
    :type model_version: str
    :return: :py:class:`Flask.Response` -- model properties
    """
    validate_model_id(model_id, model_version)

    model = app.config['model']
    model.properties.emit_update_signal()

    return jsonify(status=True)


def build_sitemap():
    """
    Build list of valid application URLs

    :return: list[str] -- list of urls
    """
    non_endpoint = [
        url.format(model_id=app.config['model'].model_id,
                   model_version=app.config['model'].model_version)
        for url in ALL_URLS
        if '{endpoint}' not in url
    ]
    with_endpoint = [
        [
            url.format(model_id=app.config['model'].model_id,
                       model_version=app.config['model'].model_version,
                       endpoint=endpoint)
            for url in ALL_URLS
            if '{endpoint}' in url
        ]
        for endpoint in app.config['model'].endpoints.keys()
    ]
    return non_endpoint + list(itertools.chain(*with_endpoint))


def page_not_found_handler(e):
    """
    Exception handler for page not found error

    :param e: NotFound exception
    :type e: :py:class:`werkzeug.exceptions.NotFound`
    :return: tuple[str, int] -- response with error code
    """
    return jsonify(error=True,
                   message=e.description,
                   valid_urls=build_sitemap()), e.code


def init_model(application):
    """
    Initialize model from app configuration

    :param application: Flask app
    :type application: :py:class:`Flask.app`
    :return: None
    """
    if 'MODEL_FILE' not in application.config:
        raise Exception('No model file provided')

    # Select model binary file path
    model_file_path = application.config['MODEL_FILE']
    LOGGER.info('Loading model from {}'.format(model_file_path))

    # Load model container
    model_container = legion.pymodel.Model.load(model_file_path)
    application.config['model'] = model_container
    LOGGER.info('Model container has been initialized')

    # Load model endpoints
    endpoints = model_container.endpoints  # force endpoints loading
    LOGGER.info('Loaded endpoints: {}'.format(list(endpoints.keys())))

    # Load model properties
    LOGGER.info('Setting properties to loaded properties {!r} (id: {})'.format(
        model_container.properties,
        id(model_container.properties)
    ))
    legion.model.set_properties(model_container.properties)

    # Force reload if code run in a cluster and model required any properties
    if legion.k8s.utils.is_code_run_in_cluster() and model_container.required_props:
        legion.model.properties.load()
        legion.model.properties.start_update_watcher()
    else:
        LOGGER.info('Ignoring running of update watcher because model is not ran in a cluster '
                    'or model does not contain required props')


def create_application():
    """
    Create Flask application and register blueprints

    :return: :py:class:`Flask.app` -- Flask application instance
    """
    static_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'static'))
    application = Flask(__name__, static_url_path='', static_path=static_folder)

    application.register_blueprint(blueprint)

    return application


def init_application(args=None):
    """
    Initialize configured Flask application instance
    Overall configuration priority: config_default.py, env::FLASK_APP_SETTINGS_FILES file,
    ENV parameters, CLI parameters

    :param args: arguments if provided
    :type args: :py:class:`argparse.Namespace` or None
    :return: :py:class:`Flask.app` -- application instance
    """
    application = create_application()
    legion.http.configure_application(application, args)

    # Put a model object into application configuration
    init_model(application)
    application.register_error_handler(404, page_not_found_handler)

    return application


def serve_model(args):
    """
    Serve model

    :param args: arguments
    :type args: :py:class:`argparse.Namespace`
    :return: None
    """
    logging.info('Legion pyserve initializing')
    application = init_application(args)

    application.run(host=application.config['LEGION_ADDR'],
                    port=application.config['LEGION_PORT'],
                    debug=application.config['DEBUG'],
                    use_reloader=False)

    return application



#
#    Copyright 2018 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Python model
"""
import json
import sys
import os
import time
import zipfile

import logging
import legion.config
import legion.containers.headers
import legion.model
import legion.k8s.properties
import legion.model.types
import legion.metrics
from legion.utils import model_properties_storage_name, send_header_to_stderr, \
    extract_archive_item, TemporaryFolder, deduce_model_file_name, save_file


import dill


LOGGER = logging.getLogger(__name__)


class ModelEndpoint:
    """
    Object describes named model invocation endpoint
    """

    def __init__(self, *, name, apply, column_types, prepare, use_df):
        """
        Build model endpoint

        :param name: endpoint name
        :type name: str
        :param apply: apply function (callable object)
        :type apply: callable
        :param column_types: (Optional) column types dictionary
        :type column_types: dict[str, ColumnInformation]
        :param prepare: (Optional) prepare function (callable object)
        :type prepare: callable
        :param use_df: use db for prepare function
        :type use_df: bool
        """
        self._name = name
        self._apply = apply
        self._column_types = column_types
        self._prepare = prepare
        self._use_df = use_df

    @property
    def description(self):
        """
        Get model description

        :return: dict[str, any] -- dict with model description
        """
        data = {
            'name': self.name,
            'use_df': self.use_df
        }

        if self.column_types:
            data['input_params'] = {k: v.description_for_api for (k, v) in self.column_types.items()}
        else:
            data['input_params'] = False

        return data

    def invoke(self, input_vector):
        """
        Calculate result of model execution

        :param input_vector: input data
        :type input_vector: dict[str, union[str, Image]]
        :return: dict -- output data
        """
        LOGGER.info('Input vector: %r' % input_vector)
        data_frame = legion.model.types.build_df(self.column_types, input_vector, not self.use_df)

        LOGGER.info('Running prepare with DataFrame: %r' % data_frame)
        data_frame = self.prepare(data_frame)

        LOGGER.info('Applying function with DataFrame: %s' % str(data_frame))
        response = self.apply(data_frame)
        LOGGER.info('Returning response: %s' % str(response))

        return response

    @property
    def name(self):
        """
        Get endpoint name

        :return: str -- name of endpoint
        """
        return self._name

    @property
    def apply(self):
        """
        Get apply function

        :return: function
        """
        return self._apply

    @property
    def column_types(self):
        """
        Get column types

        :return:
        """
        return self._column_types

    @property
    def prepare(self):
        """
        Get prepare function

        :return:
        """
        return self._prepare

    @property
    def use_df(self):
        """
        Use pandas DataFrame for input parameters

        :return: bool -- use DataFrame
        """
        return self._use_df

    def __str__(self):
        """
        Get string representation

        :return: str -- string representation
        """
        return 'ModelEndpoint for endpoint {}'.format(self.name)

    __repr__ = __str__


class Model:
    NAME = 'pymodel'

    ZIP_COMPRESSION = zipfile.ZIP_STORED
    ZIP_FILE_MODEL = 'model'
    ZIP_FILE_INFO = 'manifest.json'
    ZIP_FILE_PROPERTIES = 'properties'
    ZIP_FILE_CALLBACK = 'callback'

    PROPERTY_MODEL_ID = 'model.id'
    PROPERTY_MODEL_VERSION = 'model.version'
    PROPERTY_ENDPOINT_NAMES = 'model.endpoints'
    PROPERTY_REQUIRED_PROPERTIES = 'model.required_properties'
    PROPERTY_PROPERTIES_CALLBACK_EXISTS = 'mode.properties_callback_exists'

    def __init__(self, model_id, model_version):
        """
        Build empty model container

        :param model_id: model ID
        :type model_id: str
        :param model_version: model version
        :type model_version: str
        """
        self._meta_information = {}  # type: dict

        self._model_id = model_id
        self._model_version = model_version

        self._required_properties = []

        self._endpoints = {}  # type: dict or None
        self._path = None  # type: str or None

        self._on_property_change_callback = None  # type: typing.Callable[[], None] or None
        self._on_property_change_callback_loaded = False  # type: bool

        storage_name = model_properties_storage_name(self.model_id, self.model_version)
        self._properties = legion.k8s.K8SConfigMapStorage(storage_name,
                                                          cache_ttl=legion.config.MODEL_PROPERTIES_CACHE_TTL)

        LOGGER.info('Setting properties change callback getter to local function {!r} (id: {})'.format(
            self.get_on_property_change_callback,
            id(self.get_on_property_change_callback)
        ))

        self._properties.set_change_callback(self.get_on_property_change_callback)

        send_header_to_stderr(legion.containers.headers.MODEL_ID, self.model_id)
        send_header_to_stderr(legion.containers.headers.MODEL_VERSION, self.model_version)

    @property
    def properties(self):
        """
        Get model properties

        :return: :py:class:`legion.k8s.K8SConfigMapStorage` -- model properties storage
        """
        return self._properties

    def _load_from_archive(self, path):
        """
        Populate model container with data from file after initialization

        :param path: path to model binary
        :type path: str
        :return: None
        """
        self._path = path
        self._endpoints = None
        self._on_property_change_callback_loaded = False
        LOGGER.info('Loading model from {}'.format(path))

        LOGGER.debug('Loading metadata from {}'.format(Model.ZIP_FILE_INFO))
        with extract_archive_item(path, Model.ZIP_FILE_INFO) as manifest_path:
            with open(manifest_path, 'r') as manifest_file:
                self._meta_information = json.load(manifest_file)

        self._required_properties = self._meta_information[self.PROPERTY_REQUIRED_PROPERTIES]

        LOGGER.debug('Loading properties from {}'.format(Model.ZIP_FILE_PROPERTIES))
        with extract_archive_item(path, Model.ZIP_FILE_PROPERTIES) as properties_path:
            with open(properties_path, 'r') as properties_file:
                self.properties.data = json.load(properties_file)

        LOGGER.debug('Loading has been finished')

    @staticmethod
    def load(path):
        """
        Populate model container with data from file and initialize

        :param path: path to model binary
        :type path: str
        :return: :py:class:`legion.pymodel.model.Model` -- model container
        """
        if not os.path.exists(path):
            raise Exception('File not existed: {}'.format(path))

        with extract_archive_item(path, Model.ZIP_FILE_INFO) as manifest_path:
            with open(manifest_path, 'r') as manifest_file:
                manifest_data = json.load(manifest_file)

                model_id = manifest_data[Model.PROPERTY_MODEL_ID]
                model_version = manifest_data[Model.PROPERTY_MODEL_VERSION]

        instance = Model(model_id, model_version)
        instance._load_from_archive(path)

        return instance

    @staticmethod
    def _build_endpoint_file_name(endpoint_name):
        """
        Build name of endpoint file in model binary

        :param endpoint_name: endpoint name
        :type endpoint_name: str
        :return: inner path to file
        :rtype: str
        """
        return endpoint_name

    def load_properties_change_callback(self):
        """
        Load properties change callback from binary

        :return: deserialized model properties change callback
        :rtype: :py:class:`typing.Callable[[], None]`
        """
        LOGGER.debug('Loading properties change callback')
        with extract_archive_item(self._path, self.ZIP_FILE_CALLBACK) as callback_path:
            with open(callback_path, 'rb') as callback_file:
                return dill.load(callback_file)

    def load_endpoint(self, endpoint_name):
        """
        Load endpoint from model binary

        :param endpoint_name: endpoint name
        :type endpoint_name: str
        :return: deserialized model endpoint
        :rtype: :py:class:`legion.pymodel.model.ModelEndpoint`
        """
        LOGGER.debug('Loading endpoint {}'.format(endpoint_name))
        with extract_archive_item(self._path, self._build_endpoint_file_name(endpoint_name)) as endpoint_path:
            with open(endpoint_path, 'rb') as endpoint_file:
                return dill.load(endpoint_file)

    @property
    def description(self):
        """
        Get model container description

        :return: dict[str, any] -- model description
        """
        return {
            'model_id': self.model_id,
            'model_version': self.model_version,
            'endpoints': {ep.name: ep.description for ep in self.endpoints.values()}
        }

    @property
    def endpoints(self):
        """
        Get or lazy load endpoints

        :return: dict -- current endpoints
        """
        # If endpoint is None it means that model in partial loaded state
        # (properties has been loaded, endpoints - not)
        if self._endpoints is None:
            self._endpoints = {}

            endpoint_names = self._meta_information.get(self.PROPERTY_ENDPOINT_NAMES)
            if not endpoint_names:
                raise Exception('PyModel does not contain {} field or field is empty'
                                .format(self.PROPERTY_ENDPOINT_NAMES))

            for endpoint_name in endpoint_names:
                self._endpoints[endpoint_name] = self.load_endpoint(endpoint_name)

        return self._endpoints

    def save(self, path=None):
        """
        Save model to path (or deduce path)

        :param path: (Optional) target save name
        :return: :py:class:`legion.pymodel.model.Model` -- model container
        """
        if not self.endpoints:
            raise ValueError('Cannot save empty model container (no one export function has been called)')
        properties_change_callback_exists = self.get_on_property_change_callback() is not None

        meta_information_to_save = self._meta_information.copy()
        meta_information_to_save.update(self._collect_build_info())

        # Add model id, model version, model endpoint
        meta_information_to_save[self.PROPERTY_MODEL_ID] = self.model_id
        meta_information_to_save[self.PROPERTY_MODEL_VERSION] = self.model_version
        meta_information_to_save[self.PROPERTY_ENDPOINT_NAMES] = list(self._endpoints.keys())
        meta_information_to_save[self.PROPERTY_REQUIRED_PROPERTIES] = list(self._required_properties)
        meta_information_to_save[self.PROPERTY_PROPERTIES_CALLBACK_EXISTS] = properties_change_callback_exists

        self._path = path

        file_name_has_been_deduced = False
        if not self._path:
            self._path = deduce_model_file_name(self.model_id, self.model_version)

        LOGGER.info('Saving model to {}'.format(self._path))

        # Save
        with TemporaryFolder('legion-model-save') as temp_directory:
            temp_file = os.path.join(temp_directory.path, 'result.zip')
            with zipfile.ZipFile(temp_file, 'w', self.ZIP_COMPRESSION) as stream:
                # Add manifest file
                LOGGER.debug('Saving manifest file to {}'.format(self.ZIP_FILE_INFO))
                with open(os.path.join(temp_directory.path, self.ZIP_FILE_INFO), 'w') as info_file:
                    json.dump(meta_information_to_save, info_file)
                stream.write(os.path.join(temp_directory.path, self.ZIP_FILE_INFO), self.ZIP_FILE_INFO)

                # Add current properties state
                LOGGER.debug('Saving current property values to {}'.format(self.ZIP_FILE_PROPERTIES))
                with open(os.path.join(temp_directory.path, self.ZIP_FILE_PROPERTIES), 'w') as props_file:
                    json.dump(legion.model.properties.data, props_file)
                stream.write(os.path.join(temp_directory.path, self.ZIP_FILE_PROPERTIES), self.ZIP_FILE_PROPERTIES)

                # Add callback file
                if properties_change_callback_exists:
                    LOGGER.debug('Saving property change callback to {}'.format(self.ZIP_FILE_CALLBACK))
                    with open(os.path.join(temp_directory.path, self.ZIP_FILE_CALLBACK), 'wb') as callback_file:
                        dill.dump(self.get_on_property_change_callback(), callback_file, recurse=True)
                    stream.write(os.path.join(temp_directory.path, self.ZIP_FILE_CALLBACK), self.ZIP_FILE_CALLBACK)

                # Add endpoints
                for endpoint in self.endpoints.values():
                    path = self._build_endpoint_file_name(endpoint.name)
                    LOGGER.debug('Saving endpoint {!r} to {}'.format(endpoint, path))
                    with open(os.path.join(temp_directory.path, path), 'wb') as endpoint_file:
                        dill.dump(endpoint, endpoint_file, recurse=True)
                    stream.write(os.path.join(temp_directory.path, path), path)

            result_path = save_file(temp_file, self._path)

        send_header_to_stderr(legion.containers.headers.MODEL_PATH, result_path)
        send_header_to_stderr(legion.containers.headers.SAVE_STATUS, 'OK')

        if file_name_has_been_deduced:
            print('Model has been saved to {}'.format(result_path), file=sys.stderr)

        return self

    def _export(self, apply_func, prepare_func, column_types, use_df, endpoint_name):
        """
        Export simple Pandas based model as a bundle

        :param apply_func: an apply function DF->DF
        :type apply_func: func(x) -> y
        :param prepare_func: a function to prepare input DF->DF
        :type prepare_func: func(x) -> y
        :param column_types: result of deduce_param_types or prepared column information or None
        :type column_types: dict[str, :py:class:`legion.model.types.ColumnInformation`] or None
        :param use_df: use pandas DF for prepare and apply function
        :type use_df: bool
        :param endpoint_name: name of endpoint
        :type endpoint_name: str
        :return: None
        """
        if not hasattr(apply_func, '__call__'):
            raise Exception('Provided non-callable object as apply_function')

        if column_types:
            if not isinstance(column_types, dict) \
                    or not column_types.keys() \
                    or not isinstance(list(column_types.values())[0], legion.model.types.ColumnInformation):
                raise Exception('Bad param_types / input_data_frame provided')

        if prepare_func is None:
            def prepare_func(input_dict):
                """
                Return input value (default prepare function)
                :param x: dict of values
                :return: dict of values
                """
                return input_dict

        if endpoint_name in self._endpoints:
            raise Exception('Endpoint {} already existed in endpoints'.format(endpoint_name))

        new_endpoint = ModelEndpoint(name=endpoint_name,
                                     apply=apply_func,
                                     column_types=column_types,
                                     prepare=prepare_func,
                                     use_df=use_df)

        self.endpoints[endpoint_name] = new_endpoint

    def export_df(self, apply_func, input_data_frame, *, prepare_func=None, endpoint='default'):
        """
        Export simple Pandas DF based model as a bundle

        :param apply_func: an apply function DF->DF
        :type apply_func: func(x) -> y
        :param input_data_frame: pandas DF
        :type input_data_frame: :py:class:`pandas.DataFrame`
        :param prepare_func: a function to prepare input DF->DF
        :type prepare_func: func(x) -> y
        :param endpoint: (Optional) endpoint name, default is 'default'
        :type endpoint: str
        :return: :py:class:`legion.pymodel.model.Model` -- model container
        """
        column_types = legion.model.types.get_column_types(input_data_frame)
        self._export(apply_func, prepare_func, column_types, True, endpoint)
        return self

    def export(self, apply_func, column_types, *, prepare_func=None, endpoint='default'):
        """
        Export simple parameters defined model as a bundle

        :param apply_func: an apply function DF->DF
        :type apply_func: func(x) -> y
        :param column_types: result of deduce_param_types or prepared column information
        :type column_types: dict[str, :py:class:`legion.model.types.ColumnInformation`]
        :param prepare_func: a function to prepare input DF->DF
        :type prepare_func: func(x) -> y
        :param endpoint: (Optional) endpoint name, default is 'default'
        :type endpoint: str
        :return: :py:class:`legion.pymodel.model.Model` -- model container
        """
        self._export(apply_func, prepare_func, column_types, False, endpoint)
        return self

    def export_untyped(self, apply_func, *, prepare_func=None, endpoint='default'):
        """
        Export simple untyped model as a bundle

        :param apply_func: an apply function DF->DF
        :type apply_func: func(x) -> y
        :param prepare_func: a function to prepare input DF->DF
        :type prepare_func: func(x) -> y
        :param endpoint: (Optional) endpoint name, default is 'default'
        :type endpoint: str
        :return: :py:class:`legion.pymodel.model.Model` -- model container
        """
        self._export(apply_func, prepare_func, None, False, endpoint)
        return self

    def get_on_property_change_callback(self):
        """
        Get or lazy load registered callback or empty callback

        :return: :py:class:`Callable[[], None]` -- callback function
        """
        # If property change callback loaded - return
        if not self._on_property_change_callback_loaded:
            callback_exists = self._meta_information.get(self.PROPERTY_PROPERTIES_CALLBACK_EXISTS)

            if not callback_exists:
                LOGGER.warning('Property change callback is empty - using lambda')
                self._on_property_change_callback = lambda: None
            else:
                self._on_property_change_callback = self.load_properties_change_callback()
            self._on_property_change_callback_loaded = True

        return self._on_property_change_callback

    def on_property_change(self, callable):
        """
        Set property change callback

        :param callback: callback which will be called on each property change
        :type callback: :py:class:`Callable[[], None]`
        :return: None
        """
        self._on_property_change_callback = callable
        self._on_property_change_callback_loaded = True

    @property
    def model_id(self):
        """
        Get model id

        :return: str or None -- model id
        """
        return self._model_id

    @property
    def model_version(self):
        """
        Get model version

        :return: str or None -- model version
        """
        return self._model_version

    @property
    def required_props(self):
        """
        Get model required props

        :return: list[str] -- list of required property names
        """
        return self._required_properties

    @property
    def meta_information(self):
        """
        Get copy of meta information

        :return: dict -- copy of meta information
        """
        return self._meta_information.copy()

    def _collect_build_info(self):
        """
        Get additional container properties for container

        :return: dict[str, any] -- additional container properties
        """
        return {
            'legion.version': legion.__version__,

            'jenkins.build_number': os.environ.get(*legion.config.BUILD_NUMBER),
            'jenkins.build_id': os.environ.get(*legion.config.BUILD_ID),
            'jenkins.build_tag': os.environ.get(*legion.config.BUILD_TAG),
            'jenkins.build_url': os.environ.get(*legion.config.BUILD_URL),

            'jenkins.git_commit': os.environ.get(*legion.config.GIT_COMMIT),
            'jenkins.git_branch': os.environ.get(*legion.config.GIT_BRANCH),

            'jenkins.node_name': os.environ.get(*legion.config.NODE_NAME),
            'jenkins.job_name': os.environ.get(*legion.config.JOB_NAME)
        }

    def send_metric(self, metric, value):
        """
        Send build metric value

        :param metric: metric type or metric name
        :type metric: :py:class:`legion.metrics.Metric` or str
        :param value: metric value
        :type value: float or int
        :return: None
        """
        return legion.metrics.send_metric(self.model_id, metric, value)

    def define_property(self, name, initial_value):
        """
        Define model property and set initial value

        :param name: property name
        :type name: str
        :param initial_value: initial property value
        :type initial_value: any
        :return: :py:class:`legion.pymodel.model.Model` -- model container
        """
        self._required_properties.append(name)

        if not legion.model.properties:
            raise Exception('Model properties has not been initialized')

        legion.model.properties[name] = initial_value

        return self



#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Deploy logic for legion
"""

import logging
import os
import time
import re

import docker
import docker.errors
import legion.config
import legion.containers.docker
import legion.containers.headers
import legion.k8s
import legion.external.edi
import legion.external.grafana
import legion.pymodel
import legion.model
import legion.utils
from legion.utils import Colors, ExternalFileReader

LOGGER = logging.getLogger(__name__)

INSPECT_FORMAT_COLORIZED = 'colorized'
INSPECT_FORMAT_TABULAR = 'column'
VALID_INSPECT_FORMATS = INSPECT_FORMAT_COLORIZED, INSPECT_FORMAT_TABULAR


def build_model(args):
    """
    Build model

    :param args: command arguments
    :type args: :py:class:`argparse.Namespace`
    :return: :py:class:`docker.model.Image` docker image
    """
    client = legion.containers.docker.build_docker_client(args)

    with ExternalFileReader(args.model_file) as external_reader:
        if not os.path.exists(external_reader.path):
            raise Exception('Cannot find model binary {}'.format(external_reader.path))

        container = legion.pymodel.Model.load(external_reader.path)
        model_id = container.model_id

        image_labels = legion.containers.docker.generate_docker_labels_for_image(external_reader.path, model_id, args)

        LOGGER.info('Building docker image...')
        image = legion.containers.docker.build_docker_image(
            client,
            model_id,
            external_reader.path,
            image_labels,
            args.docker_image_tag
        )

        LOGGER.info('Image has been built: {}'.format(image))

        legion.utils.send_header_to_stderr(legion.containers.headers.IMAGE_TAG_LOCAL, image.id)

        if args.push_to_registry:
            legion.containers.docker.push_image_to_registry(client, image, args.push_to_registry)

        return image


def inspect_kubernetes(args):
    """
    Inspect kubernetes

    :param args: command arguments with .namespace
    :type args: :py:class:`argparse.Namespace`
    :return: None
    """
    edi_client = legion.external.edi.build_client(args)
    LOGGER.info('Sending inspect request to {!r}'.format(edi_client))
    model_deployments = edi_client.inspect(args.model_id, args.model_version)

    data = []

    for deployment in model_deployments:
        if deployment.status == 'ok' and deployment.model_api_ok:
            line_color = Colors.OKGREEN
        elif deployment.status == 'warning':
            line_color = Colors.WARNING
        else:
            line_color = Colors.FAIL

        errors = ''

        if not deployment.model_api_ok:
            errors = 'MODEL API DOES NOT RESPOND'

        if errors:
            errors = 'ERROR: {}'.format(errors)

        data.append({
            'deployment': deployment,
            'errors': errors,
            'line_color': line_color
        })

    if args.format == INSPECT_FORMAT_COLORIZED:
        if data:
            print('%sModel deployments:%s' % (Colors.BOLD, Colors.ENDC))

            for item in data:
                arguments = (
                    item['line_color'], Colors.ENDC,
                    Colors.UNDERLINE, item['deployment'].model, Colors.ENDC,
                    item['deployment'].image, item['deployment'].version,
                    item['line_color'], item['deployment'].ready_replicas, item['deployment'].scale, item['errors'],
                    Colors.ENDC
                )
                print('%s*%s %s%s%s %s (version: %s) - %s%s / %d pods ready %s%s' % arguments)

        if not model_deployments:
            print('%s-- cannot find any model deployments --%s' % (Colors.WARNING, Colors.ENDC))

        if not data and model_deployments:
            print('%s-- cannot find any model deployments after filtering --%s' % (Colors.WARNING, Colors.ENDC))
    elif args.format == INSPECT_FORMAT_TABULAR:
        headers = 'Model ID', 'Image', 'Version', 'Ready', 'Scale', 'Errors'
        items = [[
            item['deployment'].model,
            item['deployment'].image,
            item['deployment'].version,
            str(item['deployment'].ready_replicas),
            str(item['deployment'].scale),
            item['errors'],
        ] for item in data]

        if data:
            columns_width = [max(len(val[col_idx]) for val in items) for col_idx, column in enumerate(headers)]
            columns_width = [max(columns_width[col_idx], len(column)) for col_idx, column in enumerate(headers)]

            print('|'. join('{name:{width}} '.format(name=column, width=columns_width[col_idx])
                            for col_idx, column in enumerate(headers)))
            for item in items:
                print('|'.join('{name:{width}} '.format(name=column, width=columns_width[col_idx])
                               for col_idx, column in enumerate(item)))


def get_related_model_deployments(client, affected_deployments):
    """
    Get actual status of model deployments

    :param client: EDI client
    :type client: :py:class:`legion.external.edi.EdiClient`
    :param affected_deployments: affected by main operation (e.g. deploy) model deployments
    :type affected_deployments: list[:py:class:`legion.containers.k8s.ModelDeploymentDescription`]
    :return: list[:py:class:`legion.containers.k8s.ModelDeploymentDescription`] -- actual status of model deployments
    """
    affected_deployment_ids = {
        deployment.id_and_version
        for deployment
        in affected_deployments
    }
    actual_deployments_status = client.inspect()
    LOGGER.debug('Filtering all models using affected deployment list')
    return [deploy
            for deploy in actual_deployments_status
            if deploy.id_and_version in affected_deployment_ids]


def wait_operation_finish(args, edi_client, model_deployments, wait_callback):
    """
    Wait operation to finish according command line arguments. Uses wait_callback for checking status

    :param args: command arguments with .model_id, .namespace
    :type args: :py:class:`argparse.Namespace`
    :param edi_client: EDI client instance
    :type edi_client: :py:class:`legion.external.edi.EdiClient`
    :param model_deployments: models that have been affected during operation call
    :type model_deployments: list[:py:class:`legion.containers.k8s.ModelDeploymentDescription`]
    :param wait_callback: function that will be called to ensure that operation completed (should return True)
    :type wait_callback: py:class:`Callable[[list[:py:class:`legion.containers.k8s.ModelDeploymentDescription`]],
                                            typing.Optional[bool]]`
    :return: None
    """
    if not args.no_wait:
        start = time.time()
        if args.timeout <= 0:
            raise Exception('Invalid --timeout argument: should be positive integer')

        LOGGER.debug('Starting checking cycle limited to {} s.'.format(args.timeout))

        while True:
            elapsed = time.time() - start
            if elapsed > args.timeout:
                raise Exception('Time out: operation has not been confirmed')

            LOGGER.info('Requesting actual status of affected deployments')
            affected_deployments_status = get_related_model_deployments(edi_client, model_deployments)
            LOGGER.debug('Server returned list of actual affected deployment statuses: {!r}'
                         .format(affected_deployments_status))

            result = wait_callback(affected_deployments_status)
            if result:
                LOGGER.info('Callback have confirmed completion of the operation')
                break
            else:
                LOGGER.info('Callback have not confirmed completion of the operation')

            LOGGER.debug('Sleep before next request')
            time.sleep(1)


def check_all_scaled(deployments_status, expected_scale, expected_count):
    """
    Check that all model finished scale process and now are OK

    :param deployments_status: actual deployment status
    :type deployments_status: list[:py:class:`legion.containers.k8s.ModelDeploymentDescription`]
    :param expected_scale: expected scale
    :type expected_scale: int
    :param expected_count: expected count of models
    :type expected_count: int
    :return: bool -- result of validation
    """
    # Get fully deployed models
    finally_deployed_models = [deployment
                               for deployment in deployments_status
                               if deployment.ready_replicas == expected_scale and deployment.model_api_ok]

    # Wait until all modes will be scaled
    return len(finally_deployed_models) == expected_count


def undeploy_kubernetes(args):
    """
    Undeploy model to kubernetes

    :param args: command arguments with .model_id, .namespace
    :type args: :py:class:`argparse.Namespace`
    :return: None
    """
    edi_client = legion.external.edi.build_client(args)
    LOGGER.info('Sending undeploy request to {!r}'.format(edi_client))
    model_deployments = edi_client.undeploy(args.model_id,
                                            args.grace_period,
                                            args.model_version,
                                            args.ignore_not_found)
    LOGGER.info('Server returned list of affected deployments: {!r}'.format(model_deployments))

    wait_operation_finish(args, edi_client,
                          model_deployments,
                          lambda affected_deployments_status: not affected_deployments_status)


def scale_kubernetes(args):
    """
    Scale model instances

    :param args: command arguments with .model_id, .namespace and .scale
    :type args: :py:class:`argparse.Namespace`
    :return: None
    """
    edi_client = legion.external.edi.build_client(args)
    LOGGER.info('Sending scale request to {!r}'.format(edi_client))
    model_deployments = edi_client.scale(args.model_id, args.scale, args.model_version)
    LOGGER.info('Server returned list of affected deployments: {!r}'.format(model_deployments))

    wait_operation_finish(args, edi_client,
                          model_deployments,
                          lambda affected_deployments_status: check_all_scaled(affected_deployments_status,
                                                                               args.scale,
                                                                               len(model_deployments)))


def deploy_kubernetes(args):
    """
    Deploy kubernetes model

    :param args: command arguments with .model_id, .namespace , .livenesstimeout, .readinesstimeout and .scale
    :type args: :py:class:`argparse.Namespace`
    :return: None
    """
    edi_client = legion.external.edi.build_client(args)
    LOGGER.info('Sending deploy request to {!r}'.format(edi_client))
    model_deployments = edi_client.deploy(args.image, args.scale, args.livenesstimeout, args.readinesstimeout)
    LOGGER.info('Server returned list of affected deployments: {!r}'.format(model_deployments))

    wait_operation_finish(args, edi_client,
                          model_deployments,
                          lambda affected_deployments_status: check_all_scaled(affected_deployments_status,
                                                                               args.scale,
                                                                               len(model_deployments)))
